/* SQDRserver.cpp */
/*
 This example shows how to create, open, request, query, release, and close a mutual exclusion (mutex)
 semaphore.
*/
#define  INCL_BASE
#define INCL_DOSSEMAPHORES      /* Semaphore values */
#define INCL_DOSERRORS          /* DOS Error values */
#define INCL_DOSPROCESS
 #include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <builtin.h>
#include <time.h>

#include "SQDR.hpp"
#include "SQDRclient.hpp"

#include "sg.h"

#define VERSION "0.04"

/*+---------------------------------+*/
/*| Internal function prototypes.   |*/
/*+---------------------------------+*/
int SetupSemaphore(void);
void _Optlink SQDR_cleanup(void);
void _Optlink SQDR_ClientWork(void *param);

int sgReadConfig(char *confname);
int sgCheckForReloadConfig(char *fname,int checktime);
char *MainRedir(struct Source *src, struct SquidInfo *squidInfo, char *redirBuff, int globalPid);

char * sg_fgets(char *buf, int buflen, int fd);
int sg_puts(char *buf, int fd);

int _Optlink RedirSortCmp(const void* s1, const void* s2);
int ReDir_SortIndex(void);
int RedirSearch(char *url, int *Error);
int RedirSearch_any(char *url, int *Error);
int RedirSearch_end(char *url, int *Error);
int Add_ReDir(char *url, char *redir, int nline);
int Add_ReDir_any(char *url, char *redir, int nline);
int Add_ReDir_end(char *url, char *redir, int nline);
int HandleClientQuery(char *buf, char *bufout, int nclient);

/*+--------------------------------------------------------+*/
/*| Static global variables and local constants.           |*/
/*+--------------------------------------------------------+*/
HMTX    SQDR_hmtx     = NULLHANDLE; /* Mutex semaphore handle */

class NPipe SQDR_pipe[MAX_NUM_THREADS];
struct SQDR_threads
{
   volatile int semaphore;    // ᥬ  ஢ ६ 㯠  ⠫ 
   int num;          // ᫮ ⮪
   int next_free;    // ᫥ ᢮
   int working;      // =0  , =1 ᫥  ࢮ   (饫)
   int Nclients;     // ᫮  ⮢
   int thread_id[MAX_NUM_THREADS]; // id' ⮪
};

struct SQDR_threads  SQDRthreads = { 0,0,0,0,0 };

volatile int semnumQueryInProgress=0; // ᥬ ⢮ ࠡ뢠 ᮢ
volatile int numQueryInProgress=0;    // ⢮ ࠡ뢠 ᮢ
volatile int semReconfigInProgress=0; // ᥬ ४䨣樨
/*  ⠢ semReconfigInProgress ࠡ⪠   ந (室  sleep),
   numQueryInProgress  㢥稢, ⥭ 䨣樨 稭  numQueryInProgress=0
*/


struct LogFileStat *globalErrorLog = NULL;
struct LogFile *globalLogFile = NULL;
struct LogFileStat *lastLogFileStat;
struct LogFileStat *LogFileStat;
int sig_hup = 0;
char *globalLogDir = NULL;
char *progname;

char *configFile = "SQDR/redir.rules";
char *cWord_BaseRedirUrl = "BaseRedirUrl";
char *cWord_DefaultRedir = "DefaultRedir";
char *cWord_RedirRules   = "RedirRules";
char const  *sg_strhttp="http://", *sg_strwww="www.";

#define MAX_REDIR_TABLE     1024
#define MAX_URL_BUFF        102400
#define MAX_REDIR_URL_BUFF   51200



struct Eredir2
{  char *url;
   char *redirurl;
   int line;   // 室 ப 㫥ᮢ
   int type;   // ⨯ 㫥ᮢ: 0 - ࠢ  砫, 1 - ப, 2 - ࠢ  
   int section;// ᥪ樨 (/稪//୮/䠩)
};

struct ReDir
{
   int  Nredir;     //  redir2
   int  Nredir_any; //  redir_any
   int  Nredir_end; //  redir_end
   int  NredirUrls; //  pRedirUrl
   int  L_UrlBuff;
   int  L_RedirUrl;
   FILE *fplog;     // fp -䠩

   char BaseRedirUrl[512];
   char DefaultRedir[512];
   struct Eredir2 redir2[MAX_REDIR_TABLE];    // ⠡ ૮   砫
   struct Eredir2 redir_end[MAX_REDIR_TABLE]; // ⠡ ૮   殬 ($)
   struct Eredir2 redir_any[MAX_REDIR_TABLE]; // ⠡ ப ૮

   char UrlBuff[MAX_URL_BUFF];
   char RedirUrlBuff[MAX_REDIR_URL_BUFF];
   char *pRedirUrl[MAX_REDIR_TABLE];

};

struct ReDir rd={0,0,0,0,0,NULL};



int main(int argc, char *argv[], char *envp[])
{ int i,rc,id, idd;

   atexit(SQDR_cleanup);
/* ᥬ  ⠭ ࠧ */
   rc = SetupSemaphore();
   if(rc)
     exit(rc);

   progname = argv[0];

   sgSetGlobalErrorLogFile();
   sgReadConfig(configFile);
   sgCheckForReloadConfig(configFile,0);

   sgSetGlobalErrorLogFile();
   sgLogError("SQDR server %s started", VERSION);
   rd.fplog= fopen("SQDR/redir.log","a");

   for(i=0;i<MAX_NUM_THREADS;i++) SQDRthreads.thread_id[i] = -1;

   idd = SQDRthreads.next_free;
   id = _beginthread(SQDR_ClientWork,NULL, STACK_SIZE,(void *) &idd);

   while(__lxchg(&SQDRthreads.semaphore, LOCKED)) DosSleep(1);
   SQDRthreads.num++;
   SQDRthreads.thread_id[SQDRthreads.next_free] = id;
   if(SQDRthreads.thread_id[SQDRthreads.next_free+1] == -1) SQDRthreads.next_free++;
   else
   {  for(i=0;i<MAX_NUM_THREADS;i++)
      {  if(SQDRthreads.thread_id[i] == -1) SQDRthreads.next_free = i;
      }
   }
   __lxchg(&SQDRthreads.semaphore, UNLOCKED);

/* wait for first client connect */
   do
   {  DosSleep(1000);
   } while (!SQDRthreads.working);

  /* wait for all threads exit */
   do
   {  DosSleep(1000);
   } while (SQDRthreads.Nclients);
printf("室...\n");
   DosSleep(1000);

 }

/************************************/
void _Optlink SQDR_ClientWork(void *param)
{   int i,rc, threadNum,id,idd;
    int ncmd,data,l;
    char str[512];
    char buf[MAX_BUF], bufout[MAX_BUF];

//   HAB   habThread4 = NULLHANDLE;           /* anchor block handle for thread */
//   HMQ   hmqThread4 = NULLHANDLE;           /* message queue handle           */
//   QMSG  qmsgThread4;                       /* message queue structure        */

//   habThread4 = WinInitialize( 0UL );
//   hmqThread4 = WinCreateMsgQueue( habThread4, 0UL );

    _control87(EM_UNDERFLOW,EM_UNDERFLOW);

    threadNum =  SQDRthreads.next_free;
    if(param)
           threadNum = * ((int *)param);
    printf("Start thread %i\n",threadNum);
    sgLogError("Start thread %i\n",threadNum);
    DosSleep(500);
/* */
     if(threadNum) sprintf(str,"%s%i",SQDR_BASE_PIPE_NAME,threadNum);
     else strcpy(str,SQDR_BASE_PIPE_NAME);
    SQDR_pipe[threadNum] = NPipe(str,SERVER_MODE);
    rc = SQDR_pipe[threadNum].Create();
    if(rc)
    {  printf("訡 ᮧ  %s rc=%i",str,rc);
       if(rc == ERROR_INVALID_PARAMETER)
                   printf("(H⨬ pp)");
       exit(1);

    }
    printf("Open pipe %s rc=%i\n",str,rc);
    sgLogError("Open pipe %s rc=%i",str,rc);

    rc = SQDR_pipe[threadNum].Connect();
    printf("Connect pipe=%i\n",rc);
    sgLogError("Connect pipe=%i",rc);
    rc = SQDR_pipe[threadNum].HandShake();
    printf("HandShake=%i\n",rc);
    sgLogError("HandShake=%i",rc);

/***********/
   idd = SQDRthreads.next_free;
   id = _beginthread(SQDR_ClientWork,NULL, STACK_SIZE,&idd);
   while(__lxchg(&SQDRthreads.semaphore, LOCKED)) DosSleep(1);
    SQDRthreads.Nclients++;     // ᫮  ⮢
    SQDRthreads.working = 1;    // =0  , =1 ᫥  ࢮ   (饫)
   SQDRthreads.num++;
   SQDRthreads.thread_id[SQDRthreads.next_free] = id;
   if(SQDRthreads.thread_id[SQDRthreads.next_free+1] == -1) SQDRthreads.next_free++;
   else
   {  for(i=0;i<MAX_NUM_THREADS;i++)
      {  if(SQDRthreads.thread_id[i] == -1) SQDRthreads.next_free = i;
      }
   }
   __lxchg(&SQDRthreads.semaphore, UNLOCKED);

/*               */
/* 0 -         */
/* 1 -     */
/* 2 - ࢠ      */
   do
   {
      rc = SQDR_pipe[threadNum].RecvCmdFromClient(&ncmd,&data);
      if(rc)
      {
         if(rc == -1)
         {  rc = SQDR_pipe[threadNum].QueryState();
            if(rc == ERROR_BAD_PIPE || rc == ERROR_PIPE_NOT_CONNECTED)
                                  break; //   ??
         }
          printf("Recv error=%i\n",rc);
          exit(1);
      }
//      printf("Cmd=%i,data=%i\n",ncmd,data);
      switch(ncmd)
      {
         case 1:
               rc = SQDR_pipe[threadNum].RecvDataFromClient(buf,&l, sizeof(buf));
//               printf("%s",buf);
               HandleClientQuery(buf,bufout,threadNum);
               l = strlen(bufout) + 1;
               rc=SQDR_pipe[threadNum].SendCmdToServer(1, l);
               rc=SQDR_pipe[threadNum].SendDataToServer(bufout, l);

            break;
         case 2:
//            printf("killing plaing thread");
//            KillPlaingThead();
            break;

      }

   } while(ncmd);

    SQDR_pipe[threadNum].Close();

   while(__lxchg(&SQDRthreads.semaphore, LOCKED))
                                       DosSleep(1);
   SQDRthreads.num--;
   SQDRthreads.thread_id[threadNum] = -1;
   SQDRthreads.next_free = threadNum;
   SQDRthreads.Nclients--;     // ᫮  ⮢
   __lxchg(&SQDRthreads.semaphore, UNLOCKED);

    DosSleep(500);

//   if ( hmqThread4 != NULLHANDLE )
//      WinDestroyMsgQueue( hmqThread4 );
//   if ( habThread4 != NULLHANDLE )
//      WinTerminate( habThread4 );

//  _endthread();
}




/**************************************/
void SQDR_cleanup(void)
{   int rc;
    if(rd.fplog)
    {  sgLogError("SQDR stop");
       fclose(rd.fplog);
    }
/*  ᫥ । ᢮ ᥬ */
    if(SQDR_hmtx)
    {   rc = DosReleaseMutexSem(SQDR_hmtx);        /* Relinquish ownership */
        rc = DosCloseMutexSem(SQDR_hmtx);          /* Close mutex semaphore */
    }
}

int SetupSemaphore(void)
{
 PID     pidOwner = 0;          /* PID of current mutex semaphore owner */
 TID     tidOwner = 0;          /* TID of current mutex semaphore owner */
 ULONG   ulCount  = 0;          /* Request count for the semaphore */
 APIRET  rc       = NO_ERROR;   /* Return code */

    rc = DosCreateMutexSem(SQDRmutexName,      /* Semaphore name */
                           &SQDR_hmtx, 0, FALSE);       /* Handle returned */
    if (rc != NO_ERROR)
    {
       if(rc == ERROR_DUPLICATE_NAME)
              printf("SQDR already running\n");
       else
              printf("DosCreateMutexSem error: return code = %u\n", rc);
       return 1;
     }
  printf("DosCreateMutexSem %i\n",__LINE__);
         /* This would normally be done by another unit of work */
    rc = DosOpenMutexSem(SQDRmutexName,      /* Semaphore name */
                         &SQDR_hmtx);                 /* Handle returned */
    if (rc != NO_ERROR) {
       printf("DosOpenMutexSem error: return code = %u\n", rc);
       return 1;
     }
  printf("DosOpenMutexSem %i\n",__LINE__);

    rc = DosRequestMutexSem(SQDR_hmtx,      /* Handle of semaphore */
                            (ULONG) 1000);  /* Timeout  */
    if (rc != NO_ERROR) {
       printf("DosRequestMutexSem error: return code = %u\n", rc);
       return 1;
    }
  printf("DosRequestMutexSem %i\n",__LINE__);

    rc = DosQueryMutexSem(SQDR_hmtx,         /* Handle of semaphore */
                          &pidOwner,    /* Process ID of owner */
                          &tidOwner,    /* Thread ID of owner */
                          &ulCount);    /* Count */
    if (rc != NO_ERROR) {
       printf("DosQueryMutexSem error: return code = %u\n", rc);
       return 1;
    } else  {
       printf("Semaphore owned by PID %u, TID %u.", pidOwner, tidOwner);
       printf("  Request count is %u.\n", ulCount);
    } /* endif */
  printf("DosQueryMutexSem %i\n",__LINE__);

 return NO_ERROR;
 }

/*******************************************/
/* ⠥ 䠩  㫥ᠬ */
int sgReadConfig(char *confname)
{   int i,j,l,is, is_section, nline,rc;
    FILE *fp;
static    char str[MAX_BUF], *pstr;
static    char strtmp[MAX_BUF], str1[MAX_BUF];

    fp = fopen(confname,"r");
    if(fp == NULL)
    {
       sgLogError("Can't read conf file %s",confname);
       exit(1);

    }
// base rules
    is_section = 0;
    nline = 0;
    while(fgets(str,MAX_BUF,fp))
    {  nline++;
       l = strlen(str);
       is = 0;
       for(i=0;i<l;i++)
       {  if(str[i] > 32)
          {  if(str[i] == '#' || str[i] == ';' ) is = 1;
             break;
          }
       }
       if(is)       /* it is comment */
          continue;
       pstr = strstr(str,cWord_BaseRedirUrl);
       if(pstr)
       {   pstr += strlen(cWord_BaseRedirUrl);
           pstr = strstr(pstr,"=");
           pstr++;
           sscanf( pstr,"%s",rd.BaseRedirUrl);
           is_section++;
       }
       pstr = strstr(str,cWord_DefaultRedir);
       if(pstr)
       {   pstr += strlen(cWord_DefaultRedir);
           pstr = strstr(pstr,"=");
           pstr++;
           sscanf( pstr,"%s",rd.DefaultRedir);
           is_section++;
       }
       pstr = strstr(str,cWord_RedirRules);
       if(pstr)
       {   is_section += 2;
           break;
       }

    }

    if(!is_section)
    {  sgLogError("Not found section %s in conf file %s",cWord_BaseRedirUrl,confname);
       exit(1);
    }
    if(is_section<2)
    {  sgLogError("Not found section %s in conf file %s",cWord_RedirRules,confname);
       exit(1);
    }

//redir rules
     rd.Nredir=0;
     rd.NredirUrls=0;
     rd.Nredir_any = 0;
     rd.Nredir_end = 0;
     rd.L_UrlBuff =0;
     rd.L_RedirUrl=0;

    is_section = 0;
    while(fgets(str,MAX_BUF,fp))
    {  nline++;
       l = strlen(str);
       if(l<=1)
           continue;
       is = 0;
       for(i=0;i<l;i++)
       {  if(str[i] > 32)
          {  if(str[i] == '#' || str[i] == ';' ) is = 1;
             break;
          }
       }
       if(is)       /* it is comment */
          continue;
       str1[0]=0;
       strtmp[0] = 0;
       rc = sscanf(str,"%s %s",strtmp, str1);
       if(rc <= 0)
             continue;
       if(strtmp[0])
       {  char c_strtmplast;
          c_strtmplast = strtmp[strlen(strtmp)-1];
          if(strtmp[0] == '*')  //   ப
          {
             if(str1[0])  Add_ReDir_any(strtmp, str1, nline);
              else        Add_ReDir_any(strtmp, rd.DefaultRedir,nline);
          } else if(c_strtmplast == '$') {     //    ૠ
             if(str1[0])  Add_ReDir_end(strtmp, str1, nline);
              else        Add_ReDir_end(strtmp, rd.DefaultRedir,nline);

          } else {              //   砫 ૠ
             if(str1[0])  Add_ReDir(strtmp, str1, nline);
              else        Add_ReDir(strtmp, rd.DefaultRedir,nline);
          }
       }

    }


    fclose(fp);

    ReDir_SortIndex();
    sgLogError("ReDir_SortIndex Nredir=%i,Nredir_any=%i,L_UrlBuff=%i,NredirUrls=%i,L_RedirUrl=%i",
               rd.Nredir,rd.Nredir_any,rd.L_UrlBuff,rd.NredirUrls,rd.L_RedirUrl);

    return 0;
}

int Add_ReDir_redirurl(struct ReDir *rd, char *redirurl, int mode)
{  int i,is,l1;
   is = 0;

   for(i=0;i <  rd->NredirUrls ;i++)
   {  if(!strcmp(rd->pRedirUrl[i],redirurl) )
      { is = 1;
        if(mode == 2)
        {  rd->redir_end[rd->Nredir_end].redirurl = rd->pRedirUrl[i];
        } else if(mode == 1) {
           rd->redir_any[rd->Nredir_any].redirurl = rd->pRedirUrl[i];
        } else {
           rd->redir2[rd->Nredir].redirurl = rd->pRedirUrl[i];
        }
        break;
      }
   }
   if(is) return 2;
   l1 = strlen(redirurl);
   memcpy(&rd->RedirUrlBuff[rd->L_RedirUrl],redirurl,l1+1);
   rd->pRedirUrl[rd->NredirUrls] = &rd->RedirUrlBuff[rd->L_RedirUrl];
   if(mode == 2)
   {  rd->redir_end[rd->Nredir_end].redirurl = rd->pRedirUrl[rd->NredirUrls];
   } else if(mode == 1) {
      rd->redir_any[rd->Nredir_any].redirurl = rd->pRedirUrl[rd->NredirUrls];
   } else {
      rd->redir2[rd->Nredir].redirurl = rd->pRedirUrl[rd->NredirUrls];
   }

   rd->NredirUrls++;
   rd->L_RedirUrl += l1 +1;
   return 1;
}

/*    ᯨ᮪ ᪠  砫 ૠ */
int Add_ReDir(char *url, char *redir, int nline)
{ int l0,l1;
  if(url == NULL || redir == NULL)
                       return 1;
    if(!strncmp(url,sg_strhttp, strlen(sg_strhttp)) ) // 㡨ࠥ http://
    {  url += strlen(sg_strhttp);
    }
    if(!strncmp(url,sg_strwww, strlen(sg_strwww)) )   //㡨ࠥ www.
    {  url += strlen(sg_strwww);
    }
 l0 = strlen(url);
  l1 = strlen(redir);
  if(rd.L_UrlBuff + l0 +1 >= MAX_URL_BUFF)
  {   printf("increase MAX_URL_BUFF\n");
      sgLogError("increase MAX_URL_BUFF");
      exit(1);
  }
  if(rd.L_RedirUrl + l1 +1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_URL_BUFF\n");
      sgLogError("increase MAX_REDIR_URL_BUFF");
      exit(1);
  }

  if(rd.Nredir+ 1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_TABLE\n");
      sgLogError("increase MAX_REDIR_TABLE");
      exit(1);
  }
  memcpy(&rd.UrlBuff[rd.L_UrlBuff],url,l0+1);
  rd.redir2[rd.Nredir].url = &rd.UrlBuff[rd.L_UrlBuff];
  rd.L_UrlBuff += l0 +1;
  rd.redir2[rd.Nredir].line = nline;
  rd.redir2[rd.Nredir].type = 0;
  Add_ReDir_redirurl(&rd, redir,0);
  rd.Nredir++;
  return 0;
}

/*    ᯨ᮪ ᪠   ૠ */
int Add_ReDir_end(char *url, char *redir, int nline)
{ int l0,l1,i;

  if(url == NULL || redir == NULL)
                       return 1;
  l0 = strlen(url);
  if(url[l0-1] == '$') l0--;
  l1 = strlen(redir);
  if(rd.L_UrlBuff + l0 +1 >= MAX_URL_BUFF)
  {   printf("increase MAX_URL_BUFF\n");
      sgLogError("increase MAX_URL_BUFF");
      exit(1);
  }
  if(rd.L_RedirUrl + l1 +1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_URL_BUFF\n");
      sgLogError("increase MAX_REDIR_URL_BUFF");
      exit(1);
  }

  if(rd.Nredir_end + 1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_TABLE (Nredir_end)\n");
      sgLogError("increase MAX_REDIR_TABLE (Nredir_end)");
      exit(1);
  }
  for(i=0;i<l0;i++)
  {  rd.UrlBuff[rd.L_UrlBuff+i] = url[l0-i-1];
  }
  rd.UrlBuff[rd.L_UrlBuff+l0] = 0;

  rd.redir_end[rd.Nredir_end].url = &rd.UrlBuff[rd.L_UrlBuff];
  rd.L_UrlBuff += l0 +1;
  rd.redir_end[rd.Nredir_end].line = nline;
  rd.redir_end[rd.Nredir].type = 2;
  Add_ReDir_redirurl(&rd, redir,2);
  rd.Nredir_end++;
  return 0;
}

/*    ᯨ᮪ ᪠  ப ૠ */
int Add_ReDir_any(char *url, char *redir, int nline)
{ int l0,l1;
  if(url == NULL || redir == NULL)
                       return 1;

  if(*url == '*') url++;

  l0 = strlen(url);
  l1 = strlen(redir);
  if(rd.L_UrlBuff + l0 +1 >= MAX_URL_BUFF)
  {   printf("increase MAX_URL_BUFF\n");
      sgLogError("increase MAX_URL_BUFF");
      exit(1);
  }
  if(rd.L_RedirUrl + l1 +1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_URL_BUFF\n");
      sgLogError("increase MAX_REDIR_URL_BUFF");
      exit(1);
  }

  if(rd.Nredir_any+ 1 >= MAX_REDIR_URL_BUFF)
  {   printf("increase MAX_REDIR_TABLE (for \"any\")\n");
      sgLogError("increase MAX_REDIR_TABLE (for \"any\")");
      exit(1);
  }
  memcpy(&rd.UrlBuff[rd.L_UrlBuff],url,l0+1);
  rd.redir_any[rd.Nredir_any].url = &rd.UrlBuff[rd.L_UrlBuff];
  rd.L_UrlBuff += l0 +1;
  rd.redir_any[rd.Nredir_any].line = nline;
  rd.redir_any[rd.Nredir].type = 1;
  Add_ReDir_redirurl(&rd, redir,1);
  rd.Nredir_any++;
 return 0;
}
/* ⥭ ப  ᮪ */
/*
char * sg_fgets(char *buf, int buflen, int fd)
{  int rc;
  if(fd < 0)
       return  fgets(buf, buflen, stdin);
  rc = recv(fd,buf,buflen-1,0);

  if(rc > 0)
  {  buf[rc] = 0;
     return buf;
  }
  else
     return NULL;
}
*/
/*
int sg_puts(char *buf, int fd)
{  int rc;
static  char buff[MAX_BUF];
   if(fd < 0)
   {     puts(buf);
   } else {
      strcpy(buff,buf);
      strcat(buff,"\n");

      rc = send(fd, buff, strlen(buff), 0);
//  sgLogError("send( %x)=%i =>%s<",fd,rc,buff);
      return rc;
  }
  return 0;
}
*/

/* 㭪 ࠢ struct Eredir2 */
int _Optlink RedirSortCmp(const void* s1, const void* s2)
{   struct Eredir2 * pvInd1, *pvInd2;
    int rc;

    pvInd1 =(struct Eredir2 *) s1;
    pvInd2 =(struct Eredir2 *) s2;
    rc = strcmp(pvInd1->url,pvInd2->url);
    return rc;
}
int ReDir_SortIndex(void)
{   int rc;
    qsort( (void *) rd.redir2,rd.Nredir, sizeof(struct Eredir2), RedirSortCmp);
    qsort( (void *) rd.redir_any,rd.Nredir_any, sizeof(struct Eredir2), RedirSortCmp);
    qsort( (void *) rd.redir_end,rd.Nredir_end, sizeof(struct Eredir2), RedirSortCmp);
    return 0;
}

/*    砫 ૠ    */
int RedirSearch(char *url, int *Error)
{
    int low, high, mid,n, rc,lk,l;
    struct Eredir2 key;

    *Error = 0;
    key.url = url;
    key.line=0; key.type = 0; key.section = 0;
    low  = 0;
    n = rd.Nredir;
    if(n <= 0)
            return -1;
    high = n -1;
    while(low <= high)
    {   mid = (low + high) /2;
        rc = RedirSortCmp(&rd.redir2[mid], &key);
        if(rc > 0)
               high = mid - 1;
        else
          if( rc < 0 )
               low = mid + 1;
        else
          return mid; /* found */
    }
    *Error = 1;
    rc = RedirSortCmp(&rd.redir2[mid], &key);
    if(rc > 0 && mid)
    {            mid--;
          l =  strlen(rd.redir2[mid].url);
          lk = strlen(url);
          if(lk > l)
          {  if(!strncmp(rd.redir2[mid].url,url,l) )
                                        *Error = 0;
          }
    } else
         if(rc < 0 && mid < n) {

          l =  strlen(rd.redir2[mid].url);
          lk = strlen(url);
          if(lk > l)
          {  if(!strncmp(rd.redir2[mid].url,url,l) )
                                        *Error = 0;
          }
    }
    return mid;
}

/*     ૠ    */
int RedirSearch_end(char *url, int *Error)
{
    int low, high, mid,n, rc,lk,l,i,l0;
    struct Eredir2 key;
static  char turl[MAX_BUF];

/*  砫 ⨬ url */
   l0 = strlen(url);
   for(i=0;i<l0;i++)
   {  turl[i] = url[l0-i-1];
   }
   turl[l0] = 0;

    *Error = 0;
    key.url = &turl[0];
    key.line=0; key.type = 0; key.section = 0;
    low  = 0;
    n = rd.Nredir_end;
    if(n <= 0)
         return -1;
    high = n -1;
    while(low <= high)
    {   mid = (low + high) /2;
        rc = RedirSortCmp(&rd.redir_end[mid], &key);
        if(rc > 0)
               high = mid - 1;
        else
          if( rc < 0 )
               low = mid + 1;
        else
          return mid; /* found */
    }
    *Error = 1;
//printf("MM4 mid = %i, %s %s\n",mid, rd.redir_end[mid].url,key.url);
    rc = RedirSortCmp(&rd.redir_end[mid], &key);
//printf("MM5\n");
    if(rc > 0 && mid)
    {            mid--;
          l =  strlen(rd.redir_end[mid].url);
          lk = strlen(turl);
          if(lk > l)
          {  if(!strncmp(rd.redir_end[mid].url,turl,l) )
                                        *Error = 0;
          }
    } else
         if(rc < 0 && mid < n) {

          l =  strlen(rd.redir_end[mid].url);
          lk = strlen(url);
          if(lk > l)
          {  if(!strncmp(rd.redir_end[mid].url,turl,l) )
                                        *Error = 0;
          }
    }
    return mid;
}


/*  ॡ஬  ப   ૥ */
int RedirSearch_any(char *url, int *Error)
{
    int i, rc;

    *Error = 1;

    for(i=0;i<rd.Nredir_any;i++)
    {  if(strstr(url,rd.redir_any[i].url) )
       {  *Error = 0;
          return i;
       }
    }
    return -1;
}


int HandleClientQuery(char *buf, char *bufout, int nclient)
{
   struct SquidInfo squidInfo;
   char *redirect;
   struct Source *src;
   char redirBuff[MAX_BUF];

    if(buf == NULL || bufout == NULL)
                              return -1;
    while(semReconfigInProgress) DosSleep(1);

    if( sgCheckForReloadConfig(configFile,CHECK_CONFIG_TIME) )
    {
       while(__lxchg(&semReconfigInProgress, LOCKED)) DosSleep(1); // ᥬ
        DosBeep(1000,50);  //
        while(numQueryInProgress)  DosSleep(1); //,     믮
        sgReadConfig(configFile);
       __lxchg(&semReconfigInProgress, UNLOCKED);
    }


   while(__lxchg(&semnumQueryInProgress, LOCKED)) DosSleep(1);
   numQueryInProgress++;
   __lxchg(&semnumQueryInProgress, UNLOCKED);

    *bufout = 0;

    if(parseLine(buf,&squidInfo) != 1)
    {  sgLogError("error parsing squid line: %s",buf);
       return 1;
    }

    if((redirect = MainRedir(src,&squidInfo,redirBuff,nclient)) == NULL)
    {
//      printf("%s\n",buf);
//      sg_puts("", fd);
        return 1;
    } else {
          if(squidInfo.srcDomain[0] == '\0')
          { squidInfo.srcDomain[0] = '-';
            squidInfo.srcDomain[1] = '\0';
          }
          if(squidInfo.ident[0] == '\0')
          { squidInfo.ident[0] = '-';
            squidInfo.ident[1] = '\0';
          }

         sprintf(bufout,"%s %s/%s %s %s",redirect,squidInfo.src,
                 squidInfo.srcDomain,squidInfo.ident,
                 squidInfo.method);
//         sg_puts(buf, fd);
      }

   while(__lxchg(&semnumQueryInProgress, LOCKED)) DosSleep(1);
   numQueryInProgress--;
   __lxchg(&semnumQueryInProgress, UNLOCKED);

    return 0;
}

char *MainRedir(struct Source *src, struct SquidInfo *squidInfo, char *redirBuff, int globalPid)
{   int i,rc,err;
    FILE *fp;
    char *purl, *pcrc=NULL;
    int t1,t0,dt, rcfplog;

/*
    fprintf(fp,"Prot %s,",squidInfo->protocol);
    fprintf(fp,"Domain:%s,",squidInfo->domain);
    fprintf(fp,"Dot:%i,",squidInfo->dot);
    fprintf(fp,"Url:%s,",squidInfo->url);
    fprintf(fp,"Orig:%s,",squidInfo->orig);
    fprintf(fp,"Surl:%s,",squidInfo->surl);
    fprintf(fp,"Strippedurl:%s,",squidInfo->strippedurl);
    fprintf(fp,"Port:%i,",squidInfo->port);
    fprintf(fp,"Src:%s,",squidInfo->src);
    fprintf(fp,"SrcDomain:%s,",squidInfo->srcDomain);
    fprintf(fp,"Ident:%s,",squidInfo->ident);
    fprintf(fp,"Method:%s\n",squidInfo->method);

*/
    t0 = clock();
    purl = squidInfo->url;
    if(!strncmp(purl,sg_strhttp, strlen(sg_strhttp)) )
    {  purl += strlen(sg_strhttp);
    }
    if(!strncmp(purl,sg_strwww, strlen(sg_strwww)) )
    {  purl += strlen(sg_strwww);
    }


    rc = RedirSearch(purl, &err);
/*
      fp = fopen("SQDR/redir1.log","a");
      if(fp == NULL)
      {  if(errno == EISOPEN)
         {  DosSleep(0);
            fp = fopen("SQDR/redir1.log","a");
         }
         if(fp == NULL)
               sgLogError("MainRedir fp = %p, errno=%i",fp,errno);
      }
*/
    if(err == 0)
    {
        redirBuff[0]=0;
        if(!strstr(rd.redir2[rc].redirurl,sg_strhttp) )
        {   strcpy(redirBuff, rd.BaseRedirUrl);
            strcat(redirBuff,"/");
        }
       strcat(redirBuff,rd.redir2[rc].redirurl);
       pcrc = redirBuff;

       t1 = clock();
       dt = t1-t0;
       rcfplog = fprintf(rd.fplog,"B %s  %s  %s %i (%i)\n",squidInfo->url,
                   rd.redir2[rc].redirurl,squidInfo->src,dt,rd.redir2[rc].line);
       printf("B%i",globalPid);
       if(dt)  printf("%i",dt);
       printf(" %s %s (%i)\n",squidInfo->url,
                   rd.redir2[rc].redirurl,rd.redir2[rc].line);
    } else {
       rc =  RedirSearch_any(purl, &err);
       t1 = clock();
       dt = t1-t0;
       if(rc >= 0)
       {
          redirBuff[0]=0;
          if(!strstr(rd.redir_any[rc].redirurl,sg_strhttp) )
          {   strcpy(redirBuff, rd.BaseRedirUrl);
              strcat(redirBuff,"/");
          }
         strcat(redirBuff,rd.redir_any[rc].redirurl);
         pcrc = redirBuff;

         rcfplog = fprintf(rd.fplog,"S %s  %s  %s %i (%i)\n",squidInfo->url,
                   rd.redir_any[rc].redirurl,squidInfo->src,dt,rd.redir_any[rc].line);
          printf("S%i",globalPid);
          if(dt)  printf("%i",dt);
         printf(" %s %s (%i)\n",squidInfo->url,
                   rd.redir_any[rc].redirurl,rd.redir_any[rc].line);
       } else {
           rc =  RedirSearch_end(purl, &err);
           t1 = clock();
           dt = t1-t0;
           if(rc >= 0 && err == 0)
           {   redirBuff[0]=0;
               if(!strstr(rd.redir_end[rc].redirurl,sg_strhttp) )
               { strcpy(redirBuff, rd.BaseRedirUrl);
                 strcat(redirBuff,"/");
               }
                strcat(redirBuff,rd.redir_end[rc].redirurl);
                pcrc = redirBuff;

               rcfplog = fprintf(rd.fplog,"E %s  %s  %s %i (%i)\n",squidInfo->url,
                     rd.redir_end[rc].redirurl,squidInfo->src,dt,rd.redir_end[rc].line);
                printf("E%i",globalPid);
                if(dt)  printf("%i",dt);
                printf(" %s %s (%i)\n",squidInfo->url,
                   rd.redir_end[rc].redirurl,rd.redir_end[rc].line);
           } else {
              rcfplog =  fprintf(rd.fplog,"n %s %i\n",squidInfo->url,dt);
              if(rcfplog== -1)
                       sgLogError("write to fplog failed, (fplog=%p)",rd.fplog);

              printf("n%i",globalPid);
              if(dt)  printf("%i",dt);
              printf(" %s\n",squidInfo->url);
           }
       }
    }
    fflush(stdout);
    return pcrc;
}

